// #define VERBOSE

/**
 * @file toy_multi_cube.c
 *
 * @copyright Copyright  (C)  2013 DKRZ, MPI-M
 *
 * @author Moritz Hanke <hanke@dkrz.de>
 *         Rene Redler  <rene.redler@mpimet.mpg.de>
 *
 */
/*
 * Keywords:
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *             Rene Redler <rene.redler@mpimet.mpg.de>
 * URL: https://dkrz-sw.gitlab-pages.dkrz.de/yac/
 *
 * This file is part of YAC.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are  permitted provided that the following conditions are
 * met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the DKRZ GmbH nor the names of its contributors
 * may be used to endorse or promote products derived from this software
 * without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "yac_config.h"

#if defined YAC_NETCDF_ENABLED
#include <mpi.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "utils.h"
#include "fields.h"
#include "yac_interface.h"
#include "geometry.h"
#include "vtk_output.h"
#include "generate_cubed_sphere.h"
#include "test_function.h"

/* ------------------------------------------------- */

const char * fieldName[] = {"reg2d_conserv_field_out",
                            "unstruct_conserv_field_out",
                            "unstruct_2_conserv_field_out",
                            "cube_conserv_field_out",
                            "reg2d_avg_field_out",
                            "unstruct_avg_field_out",
                            "unstruct_2_avg_field_out",
                            "cube_avg_field_out",
                            "reg2d_hcsbb_field_out",
                            "unstruct_hcsbb_field_out",
                            "unstruct_2_hcsbb_field_out",
                            "cube_hcsbb_field_out",
                            "reg2d_2nd_conserv_field_out",
                            "unstruct_2nd_conserv_field_out",
                            "unstruct_2_2nd_conserv_field_out",
                            "cube_2nd_conserv_field_out",
                            "reg2d_rbf_field_out",
                            "unstruct_rbf_field_out",
                            "unstruct_2_rbf_field_out",
                            "cube_rbf_field_out"};

int main () {

  // Initialisation of MPI

  MPI_Init (0, NULL);

  /* The initialisation phase includes the reading of the XML file */
#ifdef VERBOSE
  printf (". main: calling yac_cinit\n");
#endif

  double tic, toc, time;

  MPI_Barrier(MPI_COMM_WORLD);

  tic=MPI_Wtime();

  yac_cinit ("coupling_toy_multi.xml", "coupling.xsd");

  /* The usual component definition, here for two sequential components on the same process */

#ifdef VERBOSE
  printf (". main: calling yac_cdef_comp\n");
#endif

  int comp_id;
  char * comp_name = "CUBE";

  yac_cdef_comp ( comp_name, &comp_id );
#ifdef VERBOSE
  printf ( ". main: defined %s with local comp ID %i \n", comp_name, comp_id );
#endif

  MPI_Comm local_comm;

  yac_cget_localcomm(&local_comm, comp_id);

  int rank, size;
  MPI_Comm_rank(local_comm,&rank);
  MPI_Comm_size(local_comm,&size);

  int ierror = 0;

  int nbr_fields;
  int cell_point_id;
  int corner_point_id;

  int field_ids[20];
  int grid_id;

  unsigned n = 50;

  unsigned nbr_vertices;
  unsigned nbr_cells;
  unsigned * num_vertices_per_cell;
  unsigned * cell_to_vertex;
  double * x_vertices;
  double * y_vertices;
  double * x_cells;
  double * y_cells;

  int * cell_core_mask;
  int * corner_core_mask;
  int * global_cell_id;
  int * global_corner_id;

  generate_part_cube_grid_information(n, &nbr_vertices, &nbr_cells,
                                      &num_vertices_per_cell, &cell_to_vertex,
                                      &x_vertices, &y_vertices, &x_cells,
                                      &y_cells, &global_cell_id,
                                      &cell_core_mask, &global_corner_id,
                                      &corner_core_mask, rank, size);

  double * x_points, * y_points;

  x_points = x_vertices;
  y_points = y_vertices;

  double * x_center, * y_center;

  x_center = x_cells;
  y_center = y_cells;

  yac_cdef_grid_unstruct ( "cube_grid", nbr_vertices, nbr_cells, (int*)num_vertices_per_cell,
                           x_vertices, y_vertices, (int*)cell_to_vertex, &grid_id );

  yac_cset_global_index(global_cell_id, CELL, grid_id);
  yac_cset_core_mask(cell_core_mask, CELL, grid_id);
  yac_cset_global_index(global_corner_id, CORNER, grid_id);
  yac_cset_core_mask(corner_core_mask, CORNER, grid_id);

  yac_cdef_points_unstruct ( grid_id, nbr_cells, CELL, x_center, y_center, &cell_point_id );
  yac_cdef_points_unstruct ( grid_id, nbr_vertices, CORNER, x_points, y_points, &corner_point_id );

  /* Field definition for the component */

  nbr_fields = 0;
  for (unsigned i = 0; i < 4; ++i)
    yac_cdef_field(fieldName[i+0], comp_id, &cell_point_id, 1, &field_ids[nbr_fields++]);
  for (unsigned i = 0; i < 4; ++i)
    yac_cdef_field(fieldName[i+4], comp_id, &corner_point_id, 1, &field_ids[nbr_fields++]);
  for (unsigned i = 0; i < 4; ++i)
    yac_cdef_field(fieldName[i+8], comp_id, &cell_point_id, 1, &field_ids[nbr_fields++]);
  for (unsigned i = 0; i < 4; ++i)
    yac_cdef_field(fieldName[i+12], comp_id, &cell_point_id, 1, &field_ids[nbr_fields++]);
  for (unsigned i = 0; i < 4; ++i)
    yac_cdef_field(fieldName[i+16], comp_id, &cell_point_id, 1, &field_ids[nbr_fields++]);
  
  toc=MPI_Wtime();
  time = toc-tic;
  printf ("toy_multi_cube: Time for initialisation %f\n", time );

  /* Search. */

#ifdef VERBOSE
  printf (". main: calling yac_csearch\n");
#endif

  MPI_Barrier(MPI_COMM_WORLD);

  tic=MPI_Wtime();

  yac_csearch ( &ierror );

  toc=MPI_Wtime();
  time = toc-tic;
  printf ("toy_multi_cube: Time for search %f\n", time );

  double * conserv_in_reg2d_point_data = xmalloc(nbr_cells * sizeof(*conserv_in_reg2d_point_data));
  double * conserv_in_unstruct_point_data = xmalloc(nbr_cells * sizeof(*conserv_in_unstruct_point_data));
  double * conserv_in_unstruct_2_point_data = xmalloc(nbr_cells * sizeof(*conserv_in_unstruct_2_point_data));

  double * avg_in_reg2d_point_data = xmalloc(nbr_vertices * sizeof(*avg_in_reg2d_point_data));
  double * avg_in_unstruct_point_data = xmalloc(nbr_vertices * sizeof(*avg_in_unstruct_point_data));
  double * avg_in_unstruct_2_point_data = xmalloc(nbr_vertices * sizeof(*avg_in_unstruct_2_point_data));

  double * hcsbb_in_reg2d_point_data = xmalloc(nbr_cells * sizeof(*hcsbb_in_reg2d_point_data));
  double * hcsbb_in_unstruct_point_data = xmalloc(nbr_cells * sizeof(*hcsbb_in_unstruct_point_data));
  double * hcsbb_in_unstruct_2_point_data = xmalloc(nbr_cells * sizeof(*hcsbb_in_unstruct_2_point_data));

  double * conserv_2nd_in_reg2d_point_data = xmalloc(nbr_cells * sizeof(*conserv_2nd_in_reg2d_point_data));
  double * conserv_2nd_in_unstruct_point_data = xmalloc(nbr_cells * sizeof(*conserv_2nd_in_unstruct_point_data));
  double * conserv_2nd_in_unstruct_2_point_data = xmalloc(nbr_cells * sizeof(*conserv_2nd_in_unstruct_2_point_data));

  double * rbf_in_reg2d_point_data = xmalloc(nbr_cells * sizeof(*rbf_in_reg2d_point_data));
  double * rbf_in_unstruct_point_data = xmalloc(nbr_cells * sizeof(*rbf_in_unstruct_point_data));
  double * rbf_in_unstruct_2_point_data = xmalloc(nbr_cells * sizeof(*rbf_in_unstruct_2_point_data));

  double * conserv_in_reg2d_point_data_err = xmalloc(nbr_cells * sizeof(*conserv_in_reg2d_point_data_err));
  double * conserv_in_unstruct_point_data_err = xmalloc(nbr_cells * sizeof(*conserv_in_unstruct_point_data_err));
  double * conserv_in_unstruct_2_point_data_err = xmalloc(nbr_cells * sizeof(*conserv_in_unstruct_2_point_data_err));

  double * avg_in_reg2d_point_data_err = xmalloc(nbr_vertices * sizeof(*avg_in_reg2d_point_data_err));
  double * avg_in_unstruct_point_data_err = xmalloc(nbr_vertices * sizeof(*avg_in_unstruct_point_data_err));
  double * avg_in_unstruct_2_point_data_err = xmalloc(nbr_vertices * sizeof(*avg_in_unstruct_2_point_data_err));

  double * hcsbb_in_reg2d_point_data_err = xmalloc(nbr_cells * sizeof(*hcsbb_in_reg2d_point_data_err));
  double * hcsbb_in_unstruct_point_data_err = xmalloc(nbr_cells * sizeof(*hcsbb_in_unstruct_point_data_err));
  double * hcsbb_in_unstruct_2_point_data_err = xmalloc(nbr_cells * sizeof(*hcsbb_in_unstruct_2_point_data_err));

  double * conserv_2nd_in_reg2d_point_data_err = xmalloc(nbr_cells * sizeof(*conserv_2nd_in_reg2d_point_data_err));
  double * conserv_2nd_in_unstruct_point_data_err = xmalloc(nbr_cells * sizeof(*conserv_2nd_in_unstruct_point_data_err));
  double * conserv_2nd_in_unstruct_2_point_data_err = xmalloc(nbr_cells * sizeof(*conserv_2nd_in_unstruct_2_point_data_err));

  double * rbf_in_reg2d_point_data_err = xmalloc(nbr_cells * sizeof(*rbf_in_reg2d_point_data_err));
  double * rbf_in_unstruct_point_data_err = xmalloc(nbr_cells * sizeof(*rbf_in_unstruct_point_data_err));
  double * rbf_in_unstruct_2_point_data_err = xmalloc(nbr_cells * sizeof(*rbf_in_unstruct_2_point_data_err));

  for (unsigned i = 0; i < nbr_cells; ++i) {
    conserv_in_reg2d_point_data[i] = -10;
    conserv_in_unstruct_point_data[i] = -10;
    conserv_in_unstruct_2_point_data[i] = -10;

    hcsbb_in_reg2d_point_data[i] = -10;
    hcsbb_in_unstruct_point_data[i] = -10;
    hcsbb_in_unstruct_2_point_data[i] = -10;

    conserv_2nd_in_reg2d_point_data[i] = -10;
    conserv_2nd_in_unstruct_point_data[i] = -10;
    conserv_2nd_in_unstruct_2_point_data[i] = -10;

    rbf_in_reg2d_point_data[i] = -10;
    rbf_in_unstruct_point_data[i] = -10;
    rbf_in_unstruct_2_point_data[i] = -10;
  }

  for (unsigned i = 0; i < nbr_vertices; ++i) {
    avg_in_reg2d_point_data[i] = -10;
    avg_in_unstruct_point_data[i] = -10;
    avg_in_unstruct_2_point_data[i] = -10;
  }

  double * cell_out_point_data = xmalloc(nbr_cells * sizeof(*cell_out_point_data));
  double * corner_out_point_data = xmalloc(nbr_vertices * sizeof(*corner_out_point_data));

  int err;
  int info;

  for (unsigned i = 0; i < nbr_cells; ++i)
    cell_out_point_data[i] = test_func(x_center[i], y_center[i]);

  for (unsigned i = 0; i < nbr_vertices; ++i)
    corner_out_point_data[i] = test_func(x_points[i], y_points[i]);

  MPI_Barrier(MPI_COMM_WORLD);

  tic=MPI_Wtime();

  {
    {
      double *point_set_data[1] = {cell_out_point_data};
      double **collection_data[1] = {point_set_data};

      yac_cput(field_ids[3], 1, collection_data, &info, &err);
      yac_cput(field_ids[11], 1, collection_data, &info, &err);
      yac_cput(field_ids[15], 1, collection_data, &info, &err);
      yac_cput(field_ids[19], 1, collection_data, &info, &err);
    }
    {
      double *point_set_data[1] = {corner_out_point_data};
      double **collection_data[1] = {point_set_data};

      yac_cput(field_ids[7], 1, collection_data, &info, &err);
    }
  }

  {
    double *collection_data[1];

    collection_data[0] = conserv_in_reg2d_point_data;
    yac_cget(field_ids[0], 1, collection_data, &info, &err);
    collection_data[0] = conserv_in_unstruct_point_data;
    yac_cget(field_ids[1], 1, collection_data, &info, &err);
    collection_data[0] = conserv_in_unstruct_2_point_data;
    yac_cget(field_ids[2], 1, collection_data, &info, &err);

    collection_data[0] = avg_in_reg2d_point_data;
    yac_cget(field_ids[4], 1, collection_data, &info, &err);
    collection_data[0] = avg_in_unstruct_point_data;
    yac_cget(field_ids[5], 1, collection_data, &info, &err);
    collection_data[0] = avg_in_unstruct_2_point_data;
    yac_cget(field_ids[6], 1, collection_data, &info, &err);

    collection_data[0] = hcsbb_in_reg2d_point_data;
    yac_cget(field_ids[8], 1, collection_data, &info, &err);
    collection_data[0] = hcsbb_in_unstruct_point_data;
    yac_cget(field_ids[9], 1, collection_data, &info, &err);
    collection_data[0] = hcsbb_in_unstruct_2_point_data;
    yac_cget(field_ids[10], 1, collection_data, &info, &err);

    collection_data[0] = conserv_2nd_in_reg2d_point_data;
    yac_cget(field_ids[12], 1, collection_data, &info, &err);
    collection_data[0] = conserv_2nd_in_unstruct_point_data;
    yac_cget(field_ids[13], 1, collection_data, &info, &err);
    collection_data[0] = conserv_2nd_in_unstruct_2_point_data;
    yac_cget(field_ids[14], 1, collection_data, &info, &err);

    collection_data[0] = rbf_in_reg2d_point_data;
    yac_cget(field_ids[16], 1, collection_data, &info, &err);
    collection_data[0] = rbf_in_unstruct_point_data;
    yac_cget(field_ids[17], 1, collection_data, &info, &err);
    collection_data[0] = rbf_in_unstruct_2_point_data;
    yac_cget(field_ids[18], 1, collection_data, &info, &err);
  }

  { // compute error for hcsbb field data

    double min_hcsbb_reg2d_err = 1337;
    double min_hcsbb_unstruct_err = 1337;
    double min_hcsbb_unstruct_2_err = 1337;
    double max_hcsbb_reg2d_err = -1337;
    double max_hcsbb_unstruct_err = -1337;
    double max_hcsbb_unstruct_2_err = -1337;


    for (unsigned i = 0; i < nbr_cells; ++i) {

      double conserv_reg2d_err = 0.0;
      double conserv_unstruct_err = 0.0;
      double conserv_unstruct_2_err = 0.0;
      double hcsbb_reg2d_err = 0.0;
      double hcsbb_unstruct_err = 0.0;
      double hcsbb_unstruct_2_err = 0.0;
      double conserv_2nd_reg2d_err = 0.0;
      double conserv_2nd_unstruct_err = 0.0;
      double conserv_2nd_unstruct_2_err = 0.0;
      double rbf_reg2d_err = 0.0;
      double rbf_unstruct_err = 0.0;
      double rbf_unstruct_2_err = 0.0;

      if (cell_core_mask[i]) {

        if ((fabs(conserv_in_reg2d_point_data[i] + 6.0) > 0.0) &&
            (fabs(conserv_in_reg2d_point_data[i] + 10.0) > 0.0))
          conserv_reg2d_err =
            (conserv_in_reg2d_point_data[i] - cell_out_point_data[i]) /
            fabs(cell_out_point_data[i]);
        if ((fabs(conserv_in_unstruct_point_data[i] + 6.0) > 0.0) &&
            (fabs(conserv_in_unstruct_point_data[i] + 10.0) > 0.0))
          conserv_unstruct_err =
            (conserv_in_unstruct_point_data[i] - cell_out_point_data[i]) /
            fabs(cell_out_point_data[i]);
        if ((fabs(conserv_in_unstruct_2_point_data[i] + 6.0) > 0.0) &&
            (fabs(conserv_in_unstruct_2_point_data[i] + 10.0) > 0.0))
          conserv_unstruct_2_err =
            (conserv_in_unstruct_2_point_data[i] - cell_out_point_data[i]) /
            fabs(cell_out_point_data[i]);

        if ((fabs(hcsbb_in_reg2d_point_data[i] + 5.0) > 0.0) &&
            (fabs(hcsbb_in_reg2d_point_data[i] + 10.0) > 0.0))
          hcsbb_reg2d_err =
            (hcsbb_in_reg2d_point_data[i] - cell_out_point_data[i]) /
            fabs(cell_out_point_data[i]);
        if ((fabs(hcsbb_in_unstruct_point_data[i] + 5.0) > 0.0) &&
            (fabs(hcsbb_in_unstruct_point_data[i] + 10.0) > 0.0))
          hcsbb_unstruct_err =
            (hcsbb_in_unstruct_point_data[i] - cell_out_point_data[i]) /
            fabs(cell_out_point_data[i]);
        if ((fabs(hcsbb_in_unstruct_2_point_data[i] + 5.0) > 0.0) &&
            (fabs(hcsbb_in_unstruct_2_point_data[i] + 10.0) > 0.0))
          hcsbb_unstruct_2_err =
            (hcsbb_in_unstruct_2_point_data[i] - cell_out_point_data[i]) /
            fabs(cell_out_point_data[i]);

        min_hcsbb_reg2d_err = MIN(min_hcsbb_reg2d_err, hcsbb_reg2d_err);
        min_hcsbb_unstruct_err = MIN(min_hcsbb_unstruct_err, hcsbb_unstruct_err);
        min_hcsbb_unstruct_2_err = MIN(min_hcsbb_unstruct_2_err, hcsbb_unstruct_2_err);
        max_hcsbb_reg2d_err = MAX(max_hcsbb_reg2d_err, hcsbb_reg2d_err);
        max_hcsbb_unstruct_err = MAX(max_hcsbb_unstruct_err, hcsbb_unstruct_err);
        max_hcsbb_unstruct_2_err = MAX(max_hcsbb_unstruct_2_err, hcsbb_unstruct_2_err);

        if ((fabs(conserv_2nd_in_reg2d_point_data[i] + 6.0) > 0.0) &&
            (fabs(conserv_2nd_in_reg2d_point_data[i] + 10.0) > 0.0))
          conserv_2nd_reg2d_err =
            (conserv_2nd_in_reg2d_point_data[i] - cell_out_point_data[i]) /
            fabs(cell_out_point_data[i]);
        if ((fabs(conserv_2nd_in_unstruct_point_data[i] + 6.0) > 0.0) &&
            (fabs(conserv_2nd_in_unstruct_point_data[i] + 10.0) > 0.0))
          conserv_2nd_unstruct_err =
            (conserv_2nd_in_unstruct_point_data[i] - cell_out_point_data[i]) /
            fabs(cell_out_point_data[i]);
        if ((fabs(conserv_2nd_in_unstruct_2_point_data[i] + 6.0) > 0.0) &&
            (fabs(conserv_2nd_in_unstruct_2_point_data[i] + 10.0) > 0.0))
          conserv_2nd_unstruct_2_err =
            (conserv_2nd_in_unstruct_2_point_data[i] - cell_out_point_data[i]) /
            fabs(cell_out_point_data[i]);

        if ((fabs(rbf_in_reg2d_point_data[i] + 6.0) > 0.0) &&
            (fabs(rbf_in_reg2d_point_data[i] + 10.0) > 0.0))
          rbf_reg2d_err =
            (rbf_in_reg2d_point_data[i] - cell_out_point_data[i]) /
            fabs(cell_out_point_data[i]);
        if ((fabs(rbf_in_unstruct_point_data[i] + 6.0) > 0.0) &&
            (fabs(rbf_in_unstruct_point_data[i] + 10.0) > 0.0))
          rbf_unstruct_err =
            (rbf_in_unstruct_point_data[i] - cell_out_point_data[i]) /
            fabs(cell_out_point_data[i]);
        if ((fabs(rbf_in_unstruct_2_point_data[i] + 6.0) > 0.0) &&
            (fabs(rbf_in_unstruct_2_point_data[i] + 10.0) > 0.0))
          rbf_unstruct_2_err =
            (rbf_in_unstruct_2_point_data[i] - cell_out_point_data[i]) /
            fabs(cell_out_point_data[i]);
      }

      conserv_in_reg2d_point_data_err[i] = conserv_reg2d_err;
      conserv_in_unstruct_point_data_err[i] = conserv_unstruct_err;
      conserv_in_unstruct_2_point_data_err[i] = conserv_unstruct_2_err;
      hcsbb_in_reg2d_point_data_err[i] = hcsbb_reg2d_err;
      hcsbb_in_unstruct_point_data_err[i] = hcsbb_unstruct_err;
      hcsbb_in_unstruct_2_point_data_err[i] = hcsbb_unstruct_2_err;
      conserv_2nd_in_reg2d_point_data_err[i] = conserv_2nd_reg2d_err;
      conserv_2nd_in_unstruct_point_data_err[i] = conserv_2nd_unstruct_err;
      conserv_2nd_in_unstruct_2_point_data_err[i] = conserv_2nd_unstruct_2_err;
      rbf_in_reg2d_point_data_err[i] = rbf_reg2d_err;
      rbf_in_unstruct_point_data_err[i] = rbf_unstruct_err;
      rbf_in_unstruct_2_point_data_err[i] = rbf_unstruct_2_err;
    }

    for (unsigned i = 0; i < nbr_vertices; ++i) {

      double avg_reg2d_err = 0.0;
      double avg_unstruct_err = 0.0;
      double avg_unstruct_2_err = 0.0;

      if (corner_core_mask[i]) {
        avg_reg2d_err =
          avg_in_reg2d_point_data[i] - corner_out_point_data[i];
        avg_unstruct_err =
          avg_in_unstruct_point_data[i] - corner_out_point_data[i];
        avg_unstruct_2_err =
          avg_in_unstruct_2_point_data[i] - corner_out_point_data[i];
      }

      avg_in_reg2d_point_data_err[i] = avg_reg2d_err;
      avg_in_unstruct_point_data_err[i] = avg_unstruct_err;
      avg_in_unstruct_2_point_data_err[i] = avg_unstruct_2_err;
    }

    MPI_Reduce(
      (rank == 0)?(MPI_IN_PLACE):(&min_hcsbb_reg2d_err),
      (rank == 0)?(&min_hcsbb_reg2d_err):(MPI_IN_PLACE),
      1, MPI_DOUBLE, MPI_MIN, 0, local_comm);
    MPI_Reduce(
      (rank == 0)?(MPI_IN_PLACE):(&min_hcsbb_unstruct_err),
      (rank == 0)?(&min_hcsbb_unstruct_err):(MPI_IN_PLACE),
      1, MPI_DOUBLE, MPI_MIN, 0, local_comm);
    MPI_Reduce(
      (rank == 0)?(MPI_IN_PLACE):(&min_hcsbb_unstruct_2_err),
      (rank == 0)?(&min_hcsbb_unstruct_2_err):(MPI_IN_PLACE),
      1, MPI_DOUBLE, MPI_MIN, 0, local_comm);
    MPI_Reduce(
      (rank == 0)?(MPI_IN_PLACE):(&max_hcsbb_reg2d_err),
      (rank == 0)?(&max_hcsbb_reg2d_err):(MPI_IN_PLACE),
      1, MPI_DOUBLE, MPI_MAX, 0, local_comm);
    MPI_Reduce(
      (rank == 0)?(MPI_IN_PLACE):(&max_hcsbb_unstruct_err),
      (rank == 0)?(&max_hcsbb_unstruct_err):(MPI_IN_PLACE),
      1, MPI_DOUBLE, MPI_MAX, 0, local_comm);
    MPI_Reduce(
      (rank == 0)?(MPI_IN_PLACE):(&max_hcsbb_unstruct_2_err),
      (rank == 0)?(&max_hcsbb_unstruct_2_err):(MPI_IN_PLACE),
      1, MPI_DOUBLE, MPI_MAX, 0, local_comm);

#ifndef HCSBB_GAUSS_NNN
#define HCSBB_GAUSS_NNN (-1)
#endif
#ifndef HCSBB_D_NNN
#define HCSBB_D_NNN (-1)
#endif
#ifndef HCSBB_GAUSS_ORDER
#define HCSBB_GAUSS_ORDER (-1)
#endif

    if (rank == 0)
      printf("HCSBB_D_NNN %d HCSBB_GAUSS_NNN %d HCSBB_GAUSS_ORDER %d reg2d %e %e\n"
             "HCSBB_D_NNN %d HCSBB_GAUSS_NNN %d HCSBB_GAUSS_ORDER %d unstruct %e %e\n"
             "HCSBB_D_NNN %d HCSBB_GAUSS_NNN %d HCSBB_GAUSS_ORDER %d unstruct_2 %e %e\n",
             HCSBB_D_NNN, HCSBB_GAUSS_NNN, HCSBB_GAUSS_ORDER, min_hcsbb_reg2d_err, max_hcsbb_reg2d_err,
             HCSBB_D_NNN, HCSBB_GAUSS_NNN, HCSBB_GAUSS_ORDER, min_hcsbb_unstruct_err, max_hcsbb_unstruct_err,
             HCSBB_D_NNN, HCSBB_GAUSS_NNN, HCSBB_GAUSS_ORDER, min_hcsbb_unstruct_2_err, max_hcsbb_unstruct_2_err);
  }

  toc=MPI_Wtime();
  time = toc-tic;
  printf ("toy_multi_cube: Time for ping-pong %f\n", time );

  //----------------------------------------------------------
  // write field to vtk output file
  //----------------------------------------------------------

  char vtk_filename[32];

  sprintf(vtk_filename, "cube_out_%d.vtk", rank);

  double point_data[nbr_vertices][3];
  for (unsigned i = 0; i < nbr_vertices; ++i) {
   LLtoXYZ(x_vertices[i], y_vertices[i], point_data[i]);
  }

  VTK_FILE *vtk_file = vtk_open(vtk_filename, "cube_out");
  vtk_write_point_data(vtk_file, (double *)point_data, nbr_vertices);
  vtk_write_cell_data(vtk_file, (unsigned *)cell_to_vertex,
                      (unsigned*)num_vertices_per_cell, nbr_cells);
  vtk_write_point_scalars_int(vtk_file, corner_core_mask, nbr_vertices,
                              "corner_core_mask");
  vtk_write_point_scalars_int(vtk_file, global_corner_id, nbr_vertices,
                              "global_corner_id");
  vtk_write_cell_scalars_int(vtk_file, cell_core_mask, nbr_cells,
                             "cell_core_mask");
  vtk_write_cell_scalars_int(vtk_file, global_cell_id, nbr_cells,
                             "global_cell_id");

  vtk_write_cell_scalars_double(vtk_file, conserv_in_reg2d_point_data, nbr_cells,
                                "conserv_in_reg2d");
  vtk_write_cell_scalars_double(vtk_file, conserv_in_unstruct_point_data, nbr_cells,
                                "conserv_in_unstruct");
  vtk_write_cell_scalars_double(vtk_file, conserv_in_unstruct_2_point_data, nbr_cells,
                                "conserv_in_unstruct_2");

  vtk_write_point_scalars_double(vtk_file, avg_in_reg2d_point_data, nbr_vertices,
                                 "avg_in_reg2d");
  vtk_write_point_scalars_double(vtk_file, avg_in_unstruct_point_data, nbr_vertices,
                                 "avg_in_unstruct");
  vtk_write_point_scalars_double(vtk_file, avg_in_unstruct_2_point_data, nbr_vertices,
                                 "avg_in_unstruct_2");

  vtk_write_cell_scalars_double(vtk_file, hcsbb_in_reg2d_point_data, nbr_cells,
                                "hcsbb_in_reg2d");
  vtk_write_cell_scalars_double(vtk_file, hcsbb_in_unstruct_point_data, nbr_cells,
                                "hcsbb_in_unstruct");
  vtk_write_cell_scalars_double(vtk_file, hcsbb_in_unstruct_2_point_data, nbr_cells,
                                "hcsbb_in_unstruct_2");

  vtk_write_cell_scalars_double(vtk_file, conserv_2nd_in_reg2d_point_data, nbr_cells,
                                "2nd_conserv_in_reg2d");
  vtk_write_cell_scalars_double(vtk_file, conserv_2nd_in_unstruct_point_data, nbr_cells,
                                "2nd_conserv_in_unstruct");
  vtk_write_cell_scalars_double(vtk_file, conserv_2nd_in_unstruct_2_point_data, nbr_cells,
                                "2nd_conserv_in_unstruct_2");

  vtk_write_cell_scalars_double(vtk_file, rbf_in_reg2d_point_data, nbr_cells,
                                "rbf_in_reg2d");
  vtk_write_cell_scalars_double(vtk_file, rbf_in_unstruct_point_data, nbr_cells,
                                "rbf_in_unstruct");
  vtk_write_cell_scalars_double(vtk_file, rbf_in_unstruct_2_point_data, nbr_cells,
                                "rbf_in_unstruct_2");

  vtk_write_cell_scalars_double(vtk_file, conserv_in_reg2d_point_data_err, nbr_cells,
                                "conserv_in_reg2d_err");
  vtk_write_cell_scalars_double(vtk_file, conserv_in_unstruct_point_data_err, nbr_cells,
                                "conserv_in_unstruct_err");
  vtk_write_cell_scalars_double(vtk_file, conserv_in_unstruct_2_point_data_err, nbr_cells,
                                "conserv_in_unstruct_2_err");

  vtk_write_point_scalars_double(vtk_file, avg_in_reg2d_point_data_err, nbr_vertices,
                                 "avg_in_reg2d_err");
  vtk_write_point_scalars_double(vtk_file, avg_in_unstruct_point_data_err, nbr_vertices,
                                 "avg_in_unstruct_err");
  vtk_write_point_scalars_double(vtk_file, avg_in_unstruct_2_point_data_err, nbr_vertices,
                                 "avg_in_unstruct_2_err");

  vtk_write_cell_scalars_double(vtk_file, hcsbb_in_reg2d_point_data_err, nbr_cells,
                                "hcsbb_in_reg2d_err");
  vtk_write_cell_scalars_double(vtk_file, hcsbb_in_unstruct_point_data_err, nbr_cells,
                                "hcsbb_in_unstruct_err");
  vtk_write_cell_scalars_double(vtk_file, hcsbb_in_unstruct_2_point_data_err, nbr_cells,
                                "hcsbb_in_unstruct_2_err");

  vtk_write_cell_scalars_double(vtk_file, conserv_2nd_in_reg2d_point_data_err, nbr_cells,
                                "2nd_conserv_in_reg2d_err");
  vtk_write_cell_scalars_double(vtk_file, conserv_2nd_in_unstruct_point_data_err, nbr_cells,
                                "2nd_conserv_in_unstruct_err");
  vtk_write_cell_scalars_double(vtk_file, conserv_2nd_in_unstruct_2_point_data_err, nbr_cells,
                                "2nd_conserv_in_unstruct_2_err");

  vtk_write_cell_scalars_double(vtk_file, rbf_in_reg2d_point_data_err, nbr_cells,
                                "rbf_in_reg2d_err");
  vtk_write_cell_scalars_double(vtk_file, rbf_in_unstruct_point_data_err, nbr_cells,
                                "rbf_in_unstruct_err");
  vtk_write_cell_scalars_double(vtk_file, rbf_in_unstruct_2_point_data_err, nbr_cells,
                                "rbf_in_unstruct_2_err");

  vtk_write_cell_scalars_double(vtk_file, cell_out_point_data, nbr_cells, "cell_out");
  vtk_write_point_scalars_double(vtk_file, corner_out_point_data, nbr_vertices, "corner_out");
  vtk_close(vtk_file);

  MPI_Comm_free(&local_comm);
  yac_cfinalize();

  MPI_Finalize();

  free(conserv_in_reg2d_point_data);
  free(conserv_in_unstruct_point_data);
  free(conserv_in_unstruct_2_point_data);
  free(conserv_in_reg2d_point_data_err);
  free(conserv_in_unstruct_point_data_err);
  free(conserv_in_unstruct_2_point_data_err);
  free(avg_in_reg2d_point_data);
  free(avg_in_unstruct_point_data);
  free(avg_in_unstruct_2_point_data);
  free(avg_in_reg2d_point_data_err);
  free(avg_in_unstruct_point_data_err);
  free(avg_in_unstruct_2_point_data_err);
  free(hcsbb_in_reg2d_point_data);
  free(hcsbb_in_unstruct_point_data);
  free(hcsbb_in_unstruct_2_point_data);
  free(hcsbb_in_reg2d_point_data_err);
  free(hcsbb_in_unstruct_point_data_err);
  free(hcsbb_in_unstruct_2_point_data_err);
  free(conserv_2nd_in_reg2d_point_data);
  free(conserv_2nd_in_unstruct_point_data);
  free(conserv_2nd_in_unstruct_2_point_data);
  free(conserv_2nd_in_reg2d_point_data_err);
  free(conserv_2nd_in_unstruct_point_data_err);
  free(conserv_2nd_in_unstruct_2_point_data_err);
  free(rbf_in_reg2d_point_data);
  free(rbf_in_unstruct_point_data);
  free(rbf_in_unstruct_2_point_data);
  free(rbf_in_reg2d_point_data_err);
  free(rbf_in_unstruct_point_data_err);
  free(rbf_in_unstruct_2_point_data_err);
  free(corner_out_point_data);
  free(cell_out_point_data);
  free(cell_core_mask);
  free(corner_core_mask);
  free(global_cell_id);
  free(global_corner_id);
  free(x_vertices);
  free(y_vertices);
  free(x_cells);
  free(y_cells);
  free(num_vertices_per_cell);
  free(cell_to_vertex);

  return EXIT_SUCCESS;
}

#else
#include <stdlib.h>
#include <stdio.h>
int main () {
  printf ("Examples requires compiling with NetCDF.\n");
  return EXIT_FAILURE;
}
#endif

