/**
 * @file instance.c
 * @brief Structs and interfaces to defined YAC instances
 *
 * @copyright Copyright  (C)  2021 Moritz Hanke <hanke@dkrz.de>
 *                                 Rene Redler <rene.redler@mpimet.mpg.de>
 *
 * @version 1.0
 * @author Moritz Hanke <hanke@dkrz.de>
 *         Rene Redler <rene.redler@mpimet.mpg.de>
 */
/*
 * Keywords:
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *             Rene Redler <rene.redler@mpimet.mpg.de>
 * URL: https://dkrz-sw.gitlab-pages.dkrz.de/yac/
 *
 * This file is part of YAC.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are  permitted provided that the following conditions are
 * met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the DKRZ GmbH nor the names of its contributors
 * may be used to endorse or promote products derived from this software
 * without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef INSTANCE_H
#define INSTANCE_H

#include "grid.h"
#include "component.h"
#include "yac_mpi.h"
#include "couple_config.h"

/** \example test_instance_parallel1.c
 * This example show how to set up a YAC instance. It uses three
 * processes.
 */

/** \example test_instance_parallel2.c
 * This example show how to set up a YAC instance. It uses nine
 * processes.
 */

/** \example test_instance_parallel3.c
 * This example show how to set up a YAC instance. It uses nine
 * processes.
 */

/** \example test_instance_parallel4.c
 * This example show how to set up a YAC instance. It uses nine
 * processes.
 */

struct yac_instance;
/**
 * Constructor for a yac instance.
 * @param[in] couple_config Coupling configuration
 * @param[in] comm          MPI communicator that contains the processes of
 *                          all components that will be registered with this
 *                          yac instance
 * @param[in] group_name      name of a group of processes
 *                            (NULL is a valid argument)
 * @return yac instance
 * @remark This is collectiv for all processes in comm. Processes that are in
 *         comm, but do not wish to initialise a YAC instance can call
 *         \ref yac_instance_dummy_new instead.
 * @remark The YAC instance will generate a communicator for all processes
 *         that passed the same non-NULL group_name. This communicator can
 *         be retrived using \ref yac_instance_get_group_comm.
 */
struct yac_instance * yac_instance_new(
  struct yac_couple_config * couple_config, MPI_Comm comm,
  char const * group_name);

/**
 * Dummy constructor, which can be called instead of \ref yac_instance_new.
 * @param[in] comm
 */
void yac_instance_dummy_new(MPI_Comm comm);

/**
 * Destructor for a yac instance
 * @param[in] instance yac instance to be deleted
 */
void yac_instance_delete(struct yac_instance * instance);

/**
 * geturns a group communicator
 * @param[in] instance yac instance
 * @return group communicator
 * @remark The group communicator contains all processes that provided the same
 *         non-NULL group name to \ref yac_instance_new. In case NULL was
 *         provided as a group name, this routine will return MPI_COMM_NULL.
 */
MPI_Comm yac_instance_get_group_comm(struct yac_instance * instance);

/**
 * Definition of job start and end datetime
 * @param[in] instance       yac instance
 * @param[in] start_datetime calendar job start datetime
 * @param[in] end_datetime   calendar job end datetime
 */
void yac_instance_def_datetime(
  struct yac_instance * instance, const char * start_datetime,
  const char * end_datetime );

/**
 * query routine for the start datetime of the job
 * @param[in] instance yac instance
 * @return start datetime
 */
char * yac_instance_get_start_datetime(struct yac_instance * instance);

/**
 * query routine for the end datetime of the job
 * @param[in] instance yac instance
 * @return end datetime
 */
char * yac_instance_get_end_datetime(struct yac_instance * instance);

/**
 * Get the coupling configuration data from a yac instance
 * @param[in] instance yac instance
 * @return coupling configuration data
 */
struct yac_couple_config * yac_instance_get_couple_config(
  struct yac_instance * instance);

/**
 * Defines the components for a yac instance
 * @param[in] instance   yac instance
 * @param[in] comp_names names of components
 * @param[in] num_comps  number of entries in comp_names
 */
void yac_instance_def_components(
  struct yac_instance * instance, char const ** comp_names, size_t num_comps);

/**
 * Adds a coupling field to a yac instance
 * @param[in] instance          yac instance
 * @param[in] field_name        name of the coupling field
 * @param[in] component_index   component index
 * @param[in] grid              grid
 * @param[in] interp_fields     interpolation fields
 * @param[in] num_interp_fields number of entries in interp_fields
 * @return pointer to coupling field
 */
struct coupling_field * yac_instance_add_field(
  struct yac_instance * instance, char const * field_name,
  size_t component_index, struct yac_basic_grid * grid,
  struct interp_field * interp_fields, size_t num_interp_fields);

/**
 * Gets a component defined in a yac instance
 * @param[in] instance          yac instance
 * @param[in] component_index   component index
 * @return component
 */
struct component * yac_instance_get_component(
  struct yac_instance * instance, size_t component_index);

/**
 * Get the coupling fields defined in a yac instance
 * @param[in]  instance   yac instance
 * @param[out] fields     array containing pointers to coupling fields
 * @param[out] num_fields number of entries in fields
 */
void yac_instance_get_fields(
  struct yac_instance * instance,
  struct coupling_field *** fields, size_t * num_fields);

/**
 * reads initiates the generation of all data structures required for
 * interpolation
 * @param[in] instance yac instance
 */
void yac_setup_instance(struct yac_instance * instance);

/**
 * returns a communicator for communicating between pairs of components
 *
 * @param[in]  instance   yac instance
 * @param[in]  comp_name1 name of first component
 * @param[in]  comp_name2 name of second component
 */
MPI_Comm yac_instance_get_pair_rootcomm(
  struct yac_instance * instance,
  char const * comp_name1, char const * comp_name2 );

#endif // INSTANCE_H
