/**
 * @file test_compute_overlap_area.c
 *
 * @copyright Copyright  (C)  2013 DKRZ, MPI-M
 *
 * @author Moritz Hanke <hanke@dkrz.de>
 *         Rene Redler  <rene.redler@mpimet.mpg.de>
 *
 * Keywords:
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *             Rene Redler <rene.redler@mpimet.mpg.de>
 * URL: https://dkrz-sw.gitlab-pages.dkrz.de/yac/
 *
 * This file is part of YAC.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are  permitted provided that the following conditions are
 * met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the DKRZ GmbH nor the names of its contributors
 * may be used to endorse or promote products derived from this software
 * without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "tests.h"
#include "clipping.h"
#include "geometry.h"
#include "area.h"
#include "test_common.h"

//#define VERBOSE

int main (void) {

  enum yac_edge_type gc_edges[5] =
    {GREAT_CIRCLE_EDGE, GREAT_CIRCLE_EDGE, GREAT_CIRCLE_EDGE, GREAT_CIRCLE_EDGE, GREAT_CIRCLE_EDGE};

  double ref_area, area;

  static double const area_tolerance = 1.0e-10;

#ifdef VERBOSE
  puts ("----- calculating area -----\n");
#endif

  {
    /* set the test cells concave pentagon with a triangle


           10.0 (lon)   11.0 (lon)
           32.0 (lat)   32.0 (lat)
                  x--------x
                 /          \
                /    10.5    \
               /     31.0     \
   9.5 (lon)  /       x        \   11.5
  30.0 (lat) x                  x  30.0


   with


                    10.5 (lon)
                    31.5 (lat)
                      x
                     / \
                    /   \
           10.0    /     \   11.0
           29.5   x-------x  29.5


    */

    struct grid_cell Pentagon =
      generate_cell_deg((double[]) {10.0, 11.0, 11.5, 10.5,  9.5},
                        (double[]) {32.0, 32.0, 30.0, 31.0, 30.0},
                        gc_edges, 5);

    struct grid_cell Triangle =
      generate_cell_deg((double[]) {10.5, 11.0, 10.0},
                        (double[]) {31.5, 29.5, 29.5},
                        gc_edges, 3);

    double Pentagon_node_xyz[3];
    double Triangle_node_xyz[3];

    LLtoXYZ_deg(10.5, 31.5, Pentagon_node_xyz);
    LLtoXYZ_deg(10.5, 30.0, Triangle_node_xyz);

    size_t N = 1;

    // the reference solution for the area
    yac_compute_overlap_areas( N, &Pentagon, Triangle, &ref_area );

    // this should give a bit identical results as we do not perform any triangulation
    yac_compute_concave_overlap_areas ( N, &Pentagon, Triangle, Triangle_node_xyz, &area );
    if ((ref_area > area) != (ref_area < area))
      PUT_ERR("error compute_concave_overlap_areas, bit identical results expected.\n");

    // with triangulation of the pentagon the result should not differ that much
    yac_compute_concave_overlap_areas ( N, &Triangle, Pentagon, Pentagon_node_xyz, &area );
    if ( fabs(ref_area-area) > area_tolerance )
      PUT_ERR("error compute_concave_overlap_areas, area difference is too large.\n");

    yac_free_grid_cell(&Triangle);
    yac_free_grid_cell(&Pentagon);
  }

  {
    /* set the test cells convex square with a triangle


                -5.0      5.0
                  x--------x  5.0
                  |        |
                  |  10.5  |
                  |  31.0  |
                  x--------x -5.0

   with

                -4.0 0.0 4.0
                      x      6.0
                     / \
                    /   \
                   /     \
                  x-------x -4.0


    */

    struct grid_cell Square =
      generate_cell_deg((double[]){-5.0, 5.0, 5.0, -5.0},
                        (double[]){-5.0, -5.0, 5.0, 5.0},
                        gc_edges, 4);

    struct grid_cell Triangle =
      generate_cell_deg((double[]){-4.0, 0.0,  4.0},
                        (double[]) {-4.0, 6.0, -4.0},
                        gc_edges, 3);

    double Square_node_xyz[3];
    double Triangle_node_xyz[3];

    LLtoXYZ_deg(0.0, 0.0, Square_node_xyz);
    LLtoXYZ_deg(0.0, 0.0, Triangle_node_xyz);

    size_t N = 1;

    // the reference solution for the area
    yac_compute_overlap_areas( N, &Square, Triangle, &ref_area );

    // this should give a bit identical results as we do not perform any triangulation
    yac_compute_concave_overlap_areas ( N, &Square, Triangle, Triangle_node_xyz, &area );
    if ((ref_area > area) != (ref_area < area))
      PUT_ERR("error compute_concave_overlap_areas, bit identical results expected.\n");

    // with triangulation of the square the result should not differ that much
    yac_compute_concave_overlap_areas ( N, &Triangle, Square, Square_node_xyz, &area );
    if ( fabs(ref_area-area) > area_tolerance )
      PUT_ERR("error compute_concave_overlap_areas, area difference is too large.\n");

    yac_free_grid_cell(&Triangle);
    yac_free_grid_cell(&Square);
  }

  {
    /* set the test cells (the target node is one of the corners of the
                           concave polygon; the center of both shapes
                           is the north pole)


                   +------+
                   |      |
                   |  /\  |
                   | /  \ |
                   +/    \+

   with

                   +------+
                   |      |
                   |      |
                   |      |
                   +------+

    */

    struct grid_cell Pentagon =
      generate_cell_deg((double[]){0,0,90,180,270},
                        (double[]){90,88,88,88,88},
                        gc_edges, 5);

    struct grid_cell Square =
      generate_cell_deg((double[]){0,90,180,270},
                        (double[]){89,89,89,89},
                        gc_edges, 4);

    double target_node_xyz[3];
    LLtoXYZ_deg(0, 90, target_node_xyz);

    size_t N = 1;

    // the reference solution for the area
    {

      struct grid_cell Triangle =
        generate_cell_deg((double[]){0,0,90},
                          (double[]){90,89,89},
                          gc_edges, 3);

      ref_area = yac_huiliers_area(Square) -
                 yac_huiliers_area(Triangle);

      yac_free_grid_cell(&Triangle);
    }

    yac_compute_concave_overlap_areas (
      N, &Pentagon, Square, target_node_xyz, &area );
    if ( fabs(ref_area-area) > area_tolerance )
      PUT_ERR("error compute_concave_osverlap_areas, area difference is too large.\n");

    yac_compute_concave_overlap_areas (
      N, &Square, Pentagon, target_node_xyz, &area );
    if ( fabs(ref_area-area) > area_tolerance )
      PUT_ERR("error compute_concave_overlap_areas, area difference is too large.\n");

    yac_free_grid_cell(&Square);
    yac_free_grid_cell(&Pentagon);
  }

  return TEST_EXIT_CODE;
}

