/**
 * @file test_init_group_c.c
 *
 * @copyright Copyright  (C)  2022 DKRZ, MPI-M
 *
 * @author Moritz Hanke <hanke@dkrz.de>
 *         Rene Redler  <rene.redler@mpimet.mpg.de>
 *
 */
/*
 * Keywords:
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *             Rene Redler <rene.redler@mpimet.mpg.de>
 * URL: https://dkrz-sw.gitlab-pages.dkrz.de/yac/
 *
 * This file is part of YAC.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are  permitted provided that the following conditions are
 * met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the DKRZ GmbH nor the names of its contributors
 * may be used to endorse or promote products derived from this software
 * without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <mpi.h>

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <yaxt.h>

#include "yac_interface.h"
#include "tests.h"

int main (void) {

  char const * xml_filename = "dummy_coupling.xml";
  char const * xsd_filename = "coupling.xsd";

  MPI_Init(NULL, NULL);
  xt_initialize(MPI_COMM_WORLD);

  int global_rank, global_size;
  MPI_Comm_rank(MPI_COMM_WORLD, &global_rank);
  MPI_Comm_size(MPI_COMM_WORLD, &global_size);

  if (global_size != 6) {
    fprintf(stderr, "Wrong number of processes (should be 6)\n");
    exit(EXIT_FAILURE);
  }

  // testing of yac_cinit_group and yac_cinit_group_instance
  {
    int yac_instance_id = -1;
    int is_dummy = 0;
    int color = MPI_UNDEFINED;
    switch (global_rank) {
      default:
      case (0):
      case (1):
        // first two ranks are part of group A
        yac_cinit_group(xml_filename, xsd_filename, "group a");
        yac_cinit_group_instance(
          xml_filename, xsd_filename, "group a", &yac_instance_id);
        color = 0;
        break;
      case (2):
      case (3):
        // next two ranks are part of group B
        yac_cinit_group(xml_filename, xsd_filename, "group b");
        yac_cinit_group_instance(
          xml_filename, xsd_filename, "group b", &yac_instance_id);
        color = 1;
        break;
      case (4):
        // rank four does not provide a group
        yac_cinit_group(xml_filename, xsd_filename, NULL);
        yac_cinit_group_instance(
          xml_filename, xsd_filename, NULL, &yac_instance_id);
        break;
      case (5):
        // rank five does a dummy initialisation
        yac_cinit_dummy();
        yac_cinit_dummy(); // instance
        is_dummy = 1;
      break;
    }

    MPI_Comm ref_group_comm;
    MPI_Comm_split(MPI_COMM_WORLD, color, 0, &ref_group_comm);

    MPI_Comm group_comm, group_comm_instance;

    if (!is_dummy) {
      yac_cget_groupcomm(&group_comm);
      yac_cget_groupcomm_instance(yac_instance_id, &group_comm_instance);
    } else {
      group_comm = MPI_COMM_NULL;
      group_comm_instance = MPI_COMM_NULL;
    }

    if (ref_group_comm != MPI_COMM_NULL) {

      int result, result_instance;
      MPI_Comm_compare(ref_group_comm, group_comm, &result);
      MPI_Comm_compare(ref_group_comm, group_comm_instance, &result_instance);

      if ((result != MPI_CONGRUENT) && (result != MPI_IDENT))
        PUT_ERR("ERROR in yac_cinit_group");
      if ((result_instance != MPI_CONGRUENT) && (result_instance != MPI_IDENT))
        PUT_ERR("ERROR in yac_cinit_group_instance");

      MPI_Comm_free(&ref_group_comm);

    } else if (group_comm != MPI_COMM_NULL) {
      PUT_ERR("ERROR in yac_cinit_group");
    } else if (group_comm != MPI_COMM_NULL) {
      PUT_ERR("ERROR in yac_cinit_group_instance");
    }

    if (!is_dummy) yac_cfinalize_instance(yac_instance_id);
    yac_cfinalize();
  }

  // testing of yac_cinit_comm_group and yac_cinit_comm_group_instance
  do {

    // generate a new "world" communicator
    MPI_Comm world_comm;
    int color = (global_rank != 3)?0:MPI_UNDEFINED;
    MPI_Comm_split(MPI_COMM_WORLD, color, 0, &world_comm);

    if (global_rank == 3) break;

    int yac_instance_id;
    int world_rank;

    MPI_Comm_rank(world_comm, &world_rank);

    switch (world_rank) {
      default:
      case (0):
      case (1):
        // first two ranks are part of group A
        yac_cinit_comm_group(xml_filename, xsd_filename, world_comm, "group a");
        yac_cinit_comm_group_instance(
          xml_filename, xsd_filename, world_comm, "group a", &yac_instance_id);
        color = 0;
        break;
      case (2):
      case (3):
        // next two ranks are part of group B
        yac_cinit_comm_group(xml_filename, xsd_filename, world_comm, "group b");
        yac_cinit_comm_group_instance(
          xml_filename, xsd_filename, world_comm, "group b", &yac_instance_id);
        color = 1;
        break;
      case (4):
        // rank four does not provide a group
        yac_cinit_comm(xml_filename, xsd_filename, world_comm);
        yac_cinit_comm_instance(
          xml_filename, xsd_filename, world_comm, &yac_instance_id);
        color = MPI_UNDEFINED;
        break;
      break;
    }

    MPI_Comm ref_group_comm;
    MPI_Comm_split(world_comm, color, 0, &ref_group_comm);

    MPI_Comm group_comm, group_comm_instance;
    yac_cget_groupcomm(&group_comm);
    yac_cget_groupcomm_instance(yac_instance_id, &group_comm_instance);

    if (ref_group_comm != MPI_COMM_NULL) {

      int result, result_instance;
      MPI_Comm_compare(ref_group_comm, group_comm, &result);
      MPI_Comm_compare(ref_group_comm, group_comm_instance, &result_instance);

      if ((result != MPI_CONGRUENT) && (result != MPI_IDENT))
        PUT_ERR("ERROR in yac_cinit_comm_group");

      if ((result_instance != MPI_CONGRUENT) && (result_instance != MPI_IDENT))
        PUT_ERR("ERROR in yac_cinit_comm_group_instance");

      MPI_Comm_free(&ref_group_comm);

    } else if (group_comm != MPI_COMM_NULL) {
      PUT_ERR("ERROR in yac_cinit_comm_group");
    } else if (group_comm_instance != MPI_COMM_NULL) {
      PUT_ERR("ERROR in yac_cinit_comm_group_instance");
    }

    yac_cfinalize_instance(yac_instance_id);
    yac_cfinalize();
  } while (0);

  // int yac_instance_id = -1;
  // yac_cinit_group_instance("dummy_coupling.xml", "coupling.xsd", "group a", yac_instance_id);
  // yac_cinit_comm_group_instance("dummy_coupling.xml", "coupling.xsd", MPI_COMM_WORLD, "group a", yac_instance_id);

  xt_finalize();
  MPI_Finalize();

  return TEST_EXIT_CODE;
}
